/**
  *  \file ptrmultilist.h
  *  \brief Multiple-Iterator Pointer List
  */
#ifndef PCC_V2_PTRMULTILIST_H
#define PCC_V2_PTRMULTILIST_H

#include <algorithm>
#include <cstddef>
#include <functional>
#include "cpluslib/assert.h"

/** List of pointers with multiple iterators, basic version.
    Implements ownership semantics for a list of pointers.
    In addition, allows multiple iterators into it,
    where deletion and/or insertion invalidates neither.

    This is the underlying implementation of ptr_multi_list<T>.
    It takes its parameterisation through a function pointer.
    It can be used standalone, but normally shouldn't.
    Its interface isn't as full/conventional as the typed
    ptr_multi_list<T>'s.

    Multiple iterators can point at ptr_multi_list_base's elements.
    If one iterator is used to delete an element, other iterators
    pointing at that element will return null. */
class ptr_multi_list_base {
    struct node;
 public:
    /** Iterator. A forward iterator. */
    class iterator {
     public:
        friend class ptr_multi_list_base;

        iterator();
        iterator(node* item, iterator** pthis);
        ~iterator();
        iterator(const iterator& other);
        iterator& operator=(const iterator& other);
        bool valid() const;
        void next();
        bool operator==(const iterator& other) const;
        bool operator!=(const iterator& other) const;
        void* get() const;
        void invalidate();

     private:
        void unlink();
        void link(const iterator& other);

        node*      item;
        node*      next_item;
        mutable iterator*  next_iterator;
        mutable iterator** pthis_iterator;
    };

    explicit ptr_multi_list_base(void deleter(void*));
    ~ptr_multi_list_base();
    void clear();
    bool empty() const;
    void push_back(void* data);
    void push_front(void* data);
    void erase(const iterator& iter);
    iterator insert(const iterator& iter, void* data);
    void swap_elems(const iterator& a, const iterator& b);
    void* extract(const iterator& iter);
    iterator begin() const;
    iterator end() const;
    void* front() const;
    void* extract_front();
    std::size_t size() const;
    void invalidate_all_iterators();

 private:
    node* insert_end(void* data);
    node* insert_before(node* item, void* data);

    struct node {
        node*  next_node;
        node** pthis_node;
        void*  data;
    };
    node*     first_node;
    node**    plast_node;
    mutable iterator* first_iterator;
    void    (*deleter)(void*);
};



/** List of pointers with multiple iterators.
    Implements ownership semantics for a list of pointers.
    In addition, allows multiple iterators into it,
    where deletion and/or insertion invalidates neither.

    If one iterator is used to delete an element, other iterators
    pointing at that element will return null.

    Iterators are forward iterators.

    This list allows append at the end, but removal only at the
    beginning. */
template<typename T>
class ptr_multi_list {
 public:
    /** Iterator. */
    class iterator {
        friend class ptr_multi_list;
     public:
        iterator(ptr_multi_list_base::iterator it)
            : it(it)
            { }
        T* operator*() const
            { return static_cast<T*>(it.get()); }
        iterator& operator++()
            { it.next(); return *this; }
        iterator operator++(int)
            { iterator tmp(it); it.next(); return tmp; }
        bool operator==(const iterator& other) const
            { return it == other.it; }
        bool operator!=(const iterator& other) const
            { return it != other.it; }
     private:
        ptr_multi_list_base::iterator it;
    };

    typedef T* value_type;
    typedef iterator const_iterator;
    typedef std::size_t size_type;
    typedef std::ptrdiff_t difference_type;

    /** Constructor. Makes a blank list. */
    ptr_multi_list()
        : data(deleter)
        { }

    /** Destructor. Clears the list and deletes all elements. */
    ~ptr_multi_list()
        { }

    /** Clear the list. Deletes all elements and invalidatas all iterators. */
    void clear()
        { data.clear(); }

    /** Check emptiness. */
    bool empty() const
        { return data.empty(); }

    /** Append a newly-allocated element. If the operation fails, the element
        will be deleted before the exception is rethrown. */
    void push_back(T* t)
        { data.push_back(t); }

    /** Prepend a newly-allocated element. If the operation fails, the element
        will be deleted before the exception is rethrown. */
    void push_front(T* t)
        { data.push_front(t); }

    /** Remove and delete first element. */
    void pop_front()
        { data.erase(data.begin()); }

    /** Erase an element. */
    void erase(const iterator& it)
        { data.erase(it.it); }

    /** Insert an element.
        \param it element is inserted before this item
        \param p element
        \return iterator pointing to newly-inserted element */
    iterator insert(const iterator& it, T* p)
        { return data.insert(it.it, p); }

    /** Swap elements. Both iterators must point at undeleted elements. */
    void swap_elems(const iterator& a, const iterator& b)
        { data.swap_elems(a.it, b.it); }

    /** Extract an element. Removes it from the sequence and returns its value.
        Caller assumes responsibility. */
    T* extract(iterator it)
        { return static_cast<T*>(data.extract(it.it)); }

    /** Get iterator to first element. */
    iterator begin() const
        { return data.begin(); }

    /** Get iterator to one-past-last element. */
    iterator end() const
        { return data.end(); }

    /** Get first element. */
    T* front() const
        { return static_cast<T*>(data.front()); }

    /** Get first element and remove it from the sequence.
        Caller assumes responsibility. */
    T* extract_front()
        { return static_cast<T*>(data.extract_front()); }

    /** Get size of list (number of elements). */
    size_type size() const
        { return data.size(); }

    /** Sort. */
    void sort()
        { sort(std::less<T*>()); }

    template<typename Pred>
    void sort(Pred p);

    /** Merge. */
    void merge(ptr_multi_list& other)
        { merge(other, std::less<T*>()); }

    template<typename Pred>
    void merge(ptr_multi_list& other, Pred p);

 private:
    static void deleter(void* p)
        { delete static_cast<T*>(p); }
    ptr_multi_list_base data;
};



/******************************** Inlines ********************************/

/** Constructor. */
inline
ptr_multi_list_base::iterator::iterator()
    : item(0),
      next_item(0),
      next_iterator(0),
      pthis_iterator(0)
{ }

/** Constructor. Builds the iterator and registers it in a list.
    \param item  Item to point at
    \param pthis Place pointer to this iterator at this place */
inline
ptr_multi_list_base::iterator::iterator(node* item, iterator** pthis)
    : item(item),
      next_item(item ? item->next_node : 0),
              next_iterator(*pthis),
              pthis_iterator(pthis)
{
    *pthis_iterator = this;
    if (next_iterator) {
        next_iterator->pthis_iterator = &next_iterator;
    }
}

/** Destructor. */
inline
ptr_multi_list_base::iterator::~iterator()
{
    unlink();
}

/** Copy constructor. */
inline
ptr_multi_list_base::iterator::iterator(const iterator& other)
    : item(other.item),
      next_item(other.next_item)
{
    link(other);
}

/** Check validity. The iterator is valid if it points at an undeleted item.
    Corrolary: it is invalid if it points at a deleted item, or after the
    end of the sequence. */
inline bool
ptr_multi_list_base::iterator::valid() const
{
    return item != 0;
}

/** Advance to next item. */
inline void
ptr_multi_list_base::iterator::next()
{
    item = next_item;
    if (next_item != 0) {
        next_item = item->next_node;
    }
}

/** Compare equality. */
inline bool
ptr_multi_list_base::iterator::operator==(const iterator& other) const
{
    return other.item == item && other.next_item == next_item;
}

/** Compare inequality. */
inline bool
ptr_multi_list_base::iterator::operator!=(const iterator& other) const
{
    return !operator==(other);
}

/** Get referenced item.
    The result will be null if this iterator points at a deleted item. */
inline void*
ptr_multi_list_base::iterator::get() const
{
    return item ? item->data : 0;
}

/** Invalidate this iterator. Turns it into an unassociated end() iterator. */
inline void
ptr_multi_list_base::iterator::invalidate()
{
    this->unlink();
    item = 0;
    next_item = 0;
}

/** Constructor.
    \param deleter Function to delete elements. */
inline
ptr_multi_list_base::ptr_multi_list_base(void deleter(void*))
    : first_node(0),
      plast_node(&first_node),
      first_iterator(0),
      deleter(deleter)
{ }

/** Destructor. Clears the list. */
inline
ptr_multi_list_base::~ptr_multi_list_base()
{
    clear();
}

/** Check emptiness.
    \return true iff this list is empty */
inline bool
ptr_multi_list_base::empty() const
{
    return first_node == 0;
}

/** Append an element.
    If appending fails, deletes the element before rethrowing the exception. */
inline void
ptr_multi_list_base::push_back(void* data)
{
    insert_end(data);
}

/** Swap elements by iterator.
    Both iterators must point at undeleted elements. */
inline void
ptr_multi_list_base::swap_elems(const iterator& a, const iterator& b)
{
    std::swap(a.item->data, b.item->data);
}

/** Get iterator to first element. */
inline ptr_multi_list_base::iterator
ptr_multi_list_base::begin() const
{
    return iterator(first_node, &first_iterator);
}

/** Get iterator to one-past-last element. */
inline ptr_multi_list_base::iterator
ptr_multi_list_base::end() const
{
    return iterator();
}

/** Get first element. */
inline void*
ptr_multi_list_base::front() const
{
    return first_node->data;
}

/** Extract first element. Removes the first element and returns it. */
inline void*
ptr_multi_list_base::extract_front()
{
    return extract(begin());
}

/** Sort with predicate. A simple merge sort.
    All iterators are invalidated.
    \param p Predicate */
template<typename T>
template<typename Pred>
void
ptr_multi_list<T>::sort(Pred p)
{
    // Move every other element into another list
    ptr_multi_list<T> tmp;
    size_type n = 0;
    for (iterator i = begin(), e = end(); i != e; ++i, ++n) {
        if ((n & 1) != 0) {
            tmp.push_back(extract(i));
        }
    }

    // Sort sub-lists
    if (n > 2) {
        tmp.sort(p);
        sort(p);
    }

    // Merge
    merge(tmp, p);
}

/** Merge with predicate. Assumes this list as well as the other list to be sorted,
    and merges the two. %other will be empty afterwards. All iterators are invalidated.
    \param p Predicate
    \param other Other list */
template<typename T>
template<typename Pred>
void
ptr_multi_list<T>::merge(ptr_multi_list& other, Pred p)
{
    iterator i = begin(), e = end();
    while (!other.empty()) {
        T* elem = other.extract_front();
        while (i != e && p(*i, elem)) {
            ++i;
        }
        insert(i, elem);
    }
}

#endif
